#include<stdlib.h>
#include<errno.h>

#include"hash.h"
#include"prime.h"

void memswap(void *loc0,void *loc1,int size) {
         while(size > sizeof(unsigned))  {
             unsigned temp = *(unsigned *)loc0;
             *(unsigned *)loc0 = *(unsigned *)loc1;
             *(unsigned *)loc1 = temp;
             (unsigned *)loc0++;
             (unsigned *)loc1++;
             size -= sizeof(unsigned);
         }
         while(size-- > 0) {
             char temp = *(char *)loc0;
             *(char *)loc0 = *(char *)loc1;
             *(char *)loc1 = temp;
             (char *)loc0++;
             (char *)loc1++;
         }
         return;
    }

void * hnew(int *about,int width,float occupancy) {
            static void *new;
            int size;

            size = cprime((int)(((double)*about)/occupancy));
            if( (new = calloc(size,width)) != NULL ) *about = size;
            return new;
       }

void hdispose(void *base) {
         free(base);
     }

void *hsearch(void *base,int width,int size,unsigned long key,hash_fn hash,match_fn match) {
         int pos,found;
         void *element,*null;

         hashes = 0;
         null = NULL;
         while((pos = hash(key,size)) >= 0) {
             element = (char *)base + (pos * width);
             if(! (found = match(element))) {
                 if(null == NULL) return element;
                 memswap(element,null,width);
                 return null;
             }
             if(found < 0 && null == NULL) null = element;
         }
         return NULL;
      }

void *_hsearch(void *base,int width,int size,unsigned long key,hash_fn hash,match_fn match) {
         int pos,found;
         void *element,*prev;

         hashes = 0;
         prev = NULL;
         while((pos = hash(key,size)) >= 0) {
             element = (char *)base + (pos * width);
             if(! match(element)) {
                 if(prev == NULL) return element;
                 memswap(element,prev,width);
                 return prev;
             }
             prev = element;
         }
         return NULL;
      }

void **hresize(void **base,int width,int *size,int newsize,key_fn key,hash_fn hash,match_fn match,fail_fn fail) {
          static void *newbase;
          void *element = *base;
          int oldsize = *size;

          newbase = hnew(&newsize,width);
          if(newbase == NULL) return NULL;
          while(oldsize > 0) {
               void *newloc = hsearch(newbase,width,newsize,key(element),hash,match);
               if(newloc == NULL) {
                   if(fail(element)) {
                      errno = EHRESIZE_FAIL;
                      free(newbase);
                      return NULL;
                   }
               }
               else memcpy(newloc,element,width);
               oldsize--;
               element = (char *)element + width;
           }
           hdispose(*base);
           *base = newbase;
           *size = newsize;
           return base;
    }

void **_hresize(void **base,int width,int *size,int newsize,key_fn key,hash_fn hash,match_fn match,fail_fn fail) {
          static void *newbase;
          void *element = *base;
          int oldsize = *size;

          newbase = hnew(&newsize,width);
          if(newbase == NULL) return NULL;
          while(oldsize > 0) {
               void *newloc = _hsearch(newbase,width,newsize,key(element),hash,match);
               if(newloc == NULL) {
                   if(fail(element)) {
                      errno = EHRESIZE_FAIL;
                      free(newbase);
                      return NULL;
                   }
               }
               else memcpy(newloc,element,width);
               oldsize--;
               element = (char *)element + width;
           }
           hdispose(*base);
           *base = newbase;
           *size = newsize;
           return base;
    }

syn_t * shnew(int *size) {
            static syn_t *new;
            int actual;

            actual = cprime(*size);
            if( (new = calloc(*size,sizeof(syn_t))) != NULL ) *size = actual;
            return new;
       }

void shdispose(syn_t *base,int size,freee_fn freee) {
        syn_t curr = *base, next = NULL;
        int index;

        for(index = 0;index < size;index++) {
            curr = base[index];
            while(curr) {
               next = curr->next;
               (freee != NULL) ? freee(curr->element) : free(curr->element);
               free(curr);
               curr = next;
            }
        }
        return;
     }

shptr_t shseek(syn_t *base,int size,int seek,shptr_t ptr) {
            int index;

            switch(seek) {
                case SHSEEK_LAST : {
                                    index = size;
                                    while(((--index) >= 0) && ! base[index]);
                                    if(index < 0) return NULL;
                                    ptr->index = index;
                                    ptr->prev = NULL;
                                    ptr->curr = base[index];
                                    while(ptr->curr->next) {
                                        ptr->prev = ptr->curr;
                                        ptr->curr = ptr->curr->next;
                                    }
                                    return ptr;
                                    }
                case SHSEEK_NEXT :{
                                   ptr->prev = ptr->curr;
                                   ptr->curr = ptr->curr->next;
                                   if(ptr->curr) return ptr;
                                   index = ptr->index;
                                   break;
                                   }
                case SHSEEK_FIRST :{
                                    index = -1;
                                    break;
                                   }
                          default : return NULL;
           }
           while(1) {
                ++index;
                if(index >= size) return NULL;
                if(base[index] != NULL) {
                       ptr->index = index;
                       ptr->prev = NULL;
                       ptr->curr = base[index];
                       return ptr;
                }
          }
       }

shptr_t shsearch(syn_t *base,int size,unsigned long key,hash_fn hash,match_fn match,shptr_t ptr) {
           ptr->index = hash(key,size);
           ptr->prev = NULL;
           ptr->curr = base[ptr->index];
           while(ptr->curr) {
               if(! match(ptr->curr->element)) {
                   if(ptr->prev) {
                      ptr->prev->next = ptr->curr->next;
                      ptr->curr->next = base[ptr->index];
                      base[ptr->index] = ptr->curr;
                   }
                   return ptr;
               }
               ptr->prev = ptr->curr;
               ptr->curr = ptr->curr->next;
           }
           return NULL;
        }

shptr_t _shsearch(syn_t *base,int size,unsigned long key,hash_fn hash,match_fn match,shptr_t ptr) {
           int notfound = 1;

           ptr->index = hash(key,size);
           ptr->prev = NULL;
           ptr->curr = base[ptr->index];
           while((ptr->curr) && (notfound = match(ptr->curr->element))) {
               if(notfound == 0) return ptr;
               else
                 if(notfound > 0) break;
               ptr->prev = ptr->curr;
               ptr->curr = ptr->curr->next;
           }
           return NULL;
        }

syn_t **shresize(syn_t **base,int *size,int newsize,key_fn key,hash_fn hash) {
            static syn_t *newbase;
            syn_t curr = NULL,next = NULL;
            int index = -1,pos;

            newbase = shnew(&newsize);
            if(! newbase) return NULL;
            while(1) {
               if(! curr) {
                  if((++index) > *size) break;
                  curr = (*base)[index];
               }
               next = curr->next;
               pos = hash(key(curr->element),newsize);
               curr->next = newbase[pos];
               newbase[pos] = curr;
               curr = next;
            }
            free(*base);
            *base = newbase;
            *size = newsize;
            return base;
       }

syn_t **_shresize(syn_t **base,int *size,int newsize,key_fn key,hash_fn hash) {
            static syn_t *newbase;
            syn_t curr = NULL,prev = NULL;
            int index = -1,pos,newindex = -1;

            newbase = shnew(&newsize);
            if(! newbase) return NULL;
            while(1) {
                if(! curr) {
                    if((++index) >= *size) break;
                    curr = (*base)[index];
                    if(prev) prev->next = curr;
                }
                pos = hash(key(curr->element),newsize);
                if(pos != newindex) {
                     if(prev) prev->next = NULL;
                     prev = NULL;
                     newindex = pos;
                     newbase[pos] = curr;
                }
                prev = curr;
                curr = curr->next;
            }
            free(*base);
            *base = newbase;
            *size = newsize;
            return base;
       }
